//////////////////////////////////////////////////////////////////////
// Surface.cpp: implementation of the CSurface class.
//
// Author: Jiri Formanek
// Date:   22.7. 2002
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#define DISPLAY_API __declspec(dllexport)

#include "Common.h"
#include "Surface.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CSurface::CSurface()
{
	m_bInit			= false;
	m_bColorKey		= false;

	m_dwHeight		= 0;
	m_dwWidth		= 0;

	m_pDisplay = disGetDisplay();

	SAFE_NULL(m_lpDDSurface);
}

CSurface::~CSurface()
{
	if(m_bInit) {		//if you do not call release() before destruct object
		Release();		//call now
	}
}

/*
 *	Create surface of specified width and height, do not fill by bitmap
 *
 *  dwWidth - width of surface
 *  dwHeight - height of surface (in pixel)
 *
 *	nFlags	- additional flags
 *			- DDFLG_VIDEO_MEMORY - creates surface in video memory if available
 *			- DDFLG_SYSTEM_MEMORY - creates surface in system memory
 *
 *
 *	Returns DD_OK or ERROR_SUCCESS if success; value 1 means that surface is already created.
 */
HRESULT		CSurface::Create(DWORD dwWidth, DWORD dwHeight, UINT nFlags)
{
	HRESULT			dwReturn = 1;// 1...ERROR_ALREADYINIT;	
	DDSURFACEDESC2  ddsd;	
	//
	// Is surface already created? Is initialized display system? 
	if(m_bInit || !m_pDisplay) {
		DXTRACE("Cannot create surface because it is already cerated or the display system is not initialized.");
		return dwReturn;
	}
	//
	// Check if the input param are valid
	ASSERT(dwWidth > 0 && dwHeight > 0);
	//
	// Create a DirectDrawSurface for this bitmap
	ZeroMemory( &ddsd, sizeof(ddsd) );
	ddsd.dwSize         = sizeof(ddsd);
	ddsd.dwFlags        = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	
	m_dwWidth  = ddsd.dwWidth  = dwWidth;		
	m_dwHeight = ddsd.dwHeight = dwHeight;
	//
	// Creation of the new surface
	dwReturn = m_pDisplay->GetDirectDraw()->CreateSurface(&ddsd, &m_lpDDSurface, NULL);		
	//
	// Try to create surface in RAM if video is full
	if(dwReturn == DDERR_OUTOFVIDEOMEMORY) {
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
		dwReturn = m_pDisplay->GetDirectDraw()->CreateSurface(&ddsd, &m_lpDDSurface, NULL);			
	}
	if(DD_OK != dwReturn) {
		DXERR("Cannot create the surface due", dwReturn);						
		return dwReturn;
	}
	m_bInit = true;		
	return dwReturn;
}

/*
 *	Create surface and fill it by specified bitmap
 *	Surface has same size as bitmap.
 *
 *	strBMP - Full path of bitmap in storage file
 *	nFlags - Additional flags about surface
 *			 DDFLG_COLORKEY - Automatic set of CK on surface
 *			 DDFLG_VIDEO_MEMORY - Create surface in video memory
 *			 DDFLG_SYSTEM_MEMORY - Create surface in the system memory
 *
 *	Returns DD_OK or ERROR_SUCCESS if successful. 
 *  Returns 1 if surface is already init or strBMP
 *  is invalid.
 *
 */
HRESULT		CSurface::Create(CString strBMP, UINT nFlags)
{
	HRESULT				dwRet = 1;// = ERROR_NOINIT;
	DWORD				dwSize;
	DWORD				dwWritten;
	LPBYTE				lpData;
	HANDLE				hFile = NULL;
	BITMAP				bm;
	HBITMAP				hBitmap;
	CString				strTempFile = _T("C:\\surface.tmp");
    HDC                 hdcImage;
    HDC                 hdc = NULL;
	//
	// Is surface initialized ?	
	// Is bitmap valid?
	// Is display initialized?
	if(m_bInit || strBMP.IsEmpty() || !m_pDisplay->IsInit()) {
		DXTRACE("General error has been occured while surface was created.");
		return dwRet;
	}
	//
	// Save internal copy of bitmap path
	m_csBitmap = strBMP;
	//
	// Find file in storage
	if((dwRet = stgGetFile3(strBMP, &lpData, dwSize)) != S_OK) {
		DXERR("Cannot find specified bitmap is storage due", dwRet);
		return dwRet;
	}
	//create temporary file
	hFile = (HANDLE)CreateFile(strTempFile, 
							GENERIC_WRITE , 
							FILE_SHARE_WRITE | FILE_SHARE_READ, 
							NULL, 
							CREATE_ALWAYS, 
							FILE_ATTRIBUTE_TEMPORARY, 
							NULL);
	if(hFile == INVALID_HANDLE_VALUE) {
		dwRet = GetLastError();
		DXERR("Cannot create temporary file due", dwRet);
		delete [] lpData;
		return dwRet;
	}
	//
	// Write bitmap data to file
	WriteFile(hFile, lpData, dwSize, &dwWritten, NULL);
	if(dwWritten != dwSize) {
		dwRet = GetLastError();
		DXERR("Cannot write data to temporary file due", dwRet);
		CloseHandle((HANDLE)hFile);								
		DeleteFile(strTempFile);	
		delete [] lpData;
		return dwRet;
	}
	//
	// Close temporary file
	CloseHandle((HANDLE)hFile);								
	//
	// Get bitmap handle					
	hBitmap = (HBITMAP)LoadImage(NULL, strTempFile, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE|LR_CREATEDIBSECTION);
	if (hBitmap == NULL) {
		dwRet = GetLastError();
		DXERR("Cannot create bitmap handle due", dwRet);
		DeleteFile(strTempFile);	
		delete [] lpData;
		return dwRet;	
	}
	//
	// Get size of the bitmap
	if(!GetObject(hBitmap, sizeof(BITMAP), &bm))
	{
		dwRet = GetLastError();
		DXERR("Cannot get info about bitmap due", dwRet);
		DeleteFile(strTempFile);	
		delete [] lpData;
		return dwRet;
	}
	//
	// Create surface
	if(ERROR_SUCCESS != (dwRet = Create(bm.bmWidth, bm.bmHeight, nFlags))) {
		DXERR("Cannot create surface due", dwRet);
		DeleteFile(strTempFile);	
		delete [] lpData;
		return dwRet;		
	}
	//
	// Make sure this surface is restored.	
	Restore();
	//
	// Create DC for our bitmap	
	hdcImage = CreateCompatibleDC(NULL);
	if (!hdcImage) {
		dwRet = GetLastError();
		DXERR("Cannot create compatible DC due", dwRet);
		DeleteFile(strTempFile);	
		delete [] lpData;
		return dwRet;		
	}
	//
	// Select bitmap intoa a memoryDC so we can use it.
	SelectObject(hdcImage, hBitmap);
	//
	// Get DC to surface
	if ((dwRet = m_lpDDSurface->GetDC(&hdc)) != DD_OK)
	{
		DXERR("Cannot get surface DC due", dwRet);
		return dwRet;
	}
	//
	// Blit bitmap to GDI surface	
	if(0 == StretchBlt(hdc, 0, 0, bm.bmWidth, bm.bmHeight, hdcImage, 0, 0, bm.bmWidth , bm.bmHeight, SRCCOPY)) 
	{
		dwRet = GetLastError();
		DXERR("Cannot blit bitmap to surface due", dwRet);
		return dwRet;
	}
	//
	// Release DC
	if(DD_OK != (dwRet = m_lpDDSurface->ReleaseDC(hdc))) {
		DXERR("Cannot release surface DC due", dwRet);
		return dwRet;
	}	
	//
	// If we want to set color key, set it now
	if(nFlags & DDFLG_COLORKEY) {
		dwRet = SetColorKey();										
	}		
	//
	//release handle to GDI object and handle to DC								
	DeleteDC(hdcImage);		
	DeleteObject(hBitmap);
	DeleteFile(strTempFile);				
	delete [] lpData;	

	return dwRet;
}

/*
 *	Init CSurface object from existing surface
 *
 *	lpSurface - pointer to existing surface
 *
 *	Returns DD_OK or ERROR_SUCCESS if success;
 *	Returns 1 if pointer to surface is invalid or
 *  if object of CSurface is already created
 */
HRESULT		CSurface::Create(LPDIRECTDRAWSURFACE7 lpSurface)
{
	DDSURFACEDESC2 ddsd;
	ddsd.dwSize = sizeof(ddsd);
	HRESULT		dwReturn = 1;// = ERROR_INIT or INVALID SURFACE;	
	if(lpSurface && !m_bInit) {
		dwReturn = lpSurface->GetSurfaceDesc(&ddsd);
		if(dwReturn != DD_OK) {			
			DXERR("You may pass invalid surface. Cannot get info about due", dwReturn);
			return dwReturn;
		}
		//
		// Init member fuction according existing surface
		m_lpDDSurface  = lpSurface;
		m_dwHeight     = ddsd.dwHeight;
		m_dwWidth      = ddsd.dwWidth;
		m_bColorKey    = ddsd.dwFlags & DDSD_CKSRCBLT;
		m_bInit        = true;
		m_csBitmap     = _S_EMPTY;
	}		
	return dwReturn;
}

/*
 *	Releases created surface (is used as destructor)
 *
 *	Always returns ERROR_SUCCESS.
 *
 */
HRESULT		CSurface::Release()
{
	SAFE_RELEASE(m_lpDDSurface);			//safe release surface
	m_bInit			= false;				//and reset all member variable
	m_bColorKey		= false;
	m_dwHeight		= 0;
	m_dwWidth		= 0;	

	return ERROR_SUCCESS;
}

/*
 *	Restores lost surface
 *
 *	bFill - indicates when surfaces is refilled by original bitmap.
 *			TRUE for refilling and FALSE only for restoration.
 *
 *	Returns DD_OK or ERROR_SUCCESS if success; Returns 1 if suraface 
 *	is not created;
 *
 */
HRESULT		CSurface::Restore(BOOL	bFill)
{
	HRESULT dwReturn = 1;// 1...ERROR_NOINIT;
	if(m_bInit) {
		//
		// Try tu restore surface
		dwReturn = m_lpDDSurface->Restore();
		if(dwReturn != DD_OK) {
			DXERR("Cannot restore surface due", dwReturn);
			return dwReturn;
		}
		if(bFill) {
			//
			// Refill surface by bitmap
			dwReturn = CopyBitmap(m_csBitmap);				
		}		
	}
	return dwReturn;
}


/*
 *	Set color key on surface
 *	Transparent color is chosen from the first pixel of surface
 *
 *	Returns DD_OK if success.
 *
 */
HRESULT		CSurface::SetColorKey()
{
	HRESULT				dwReturn = 1;
	DWORD				dwPixel;
	DDSURFACEDESC2		ddsd;
	DDCOLORKEY          ddck;			

	if(m_bInit) {							
		//	
		// Now lock the surface so we can read back the converted color		
		ddsd.dwSize = sizeof(ddsd);
		dwReturn = GetSurface()->Lock( NULL, &ddsd, DDLOCK_WAIT, NULL );
		if(dwReturn != DD_OK) {
			DXERR("Cannot lock surface to set color key due", dwReturn);
			return dwReturn;
		}		
		//
		// Get first pixel
		dwPixel = *(DWORD *) ddsd.lpSurface; 
		//
		// Mask it to bpp
		if(ddsd.ddpfPixelFormat.dwRGBBitCount < 32) {
			dwPixel &= (1 << ddsd.ddpfPixelFormat.dwRGBBitCount) - 1;  
		}
		//
		// Unlock surface
		GetSurface()->Unlock(NULL);
		//
		// Set color values
		ddck.dwColorSpaceLowValue = ddck.dwColorSpaceHighValue = dwPixel;		
		//
		// Set color key on surface
		dwReturn = m_lpDDSurface->SetColorKey(DDCKEY_SRCBLT, &ddck);
		if(dwReturn != DD_OK) {
			DXERR("Cannot set color key on surface due", dwReturn);
		}		
		if(dwReturn == DD_OK) {
			m_bColorKey = true;
		}
	}
	return dwReturn;
}

/*
 *	Returns pointer to DirectDraw Surface
 *
 *  If surface is not init returns NULL 
 *
 */
LPDIRECTDRAWSURFACE7 CSurface::GetSurface()
{
	//
	// Is surface initialized ?
	if(m_bInit) {								
		return m_lpDDSurface;	
	}
	return NULL;				
}

/*
 *	Check if the surface is valid created
 *	and pointer is valid
 *
 *	Returns TRUE if valid else FALSE
 */
BOOL	CSurface::IsValid()
{
	if(m_lpDDSurface && IsInit()) {
		return TRUE;
	}
	else {
		return FALSE;
	}
}

/*
 *	Blit solid color rectangle to surface
 *
 *	Parameters:		rcDestin	- destination rectangle to blit
 *					rcSource	- source rectangle from to blit
 *					surSource	- source surface, must be valid
 *
 *	Return values:  DD_OK if success, else error code of Blit() function
 */
HRESULT CSurface::Blt(CRect rcDestin, CRect rcSource, CSurface* surSource)
{
	HRESULT dwResult = 1; // 1 means: NO INIT
	//
	// Check if the destination surface is valid
	if(IsValid() && surSource->IsValid()) {		
		//
		// Blit with color key
		if(surSource->IsColorKey()) {
			while(1) {
				dwResult = GetSurface()->Blt(&rcDestin, surSource->GetSurface(), &rcSource, DDBLT_KEYSRC, NULL);
				if(dwResult != DDERR_WASSTILLDRAWING) {
					break;
				}
			}
		}
		//
		// Blit without color key
		else {
			while(1) {
				dwResult = GetSurface()->Blt(&rcDestin, surSource->GetSurface(), &rcSource, 0, NULL);
				if(dwResult != DDERR_WASSTILLDRAWING) {
					break;
				}
			}
		}		
		if(DD_OK != dwResult) {	
			DXERR("Cannot blit solid rect due", dwResult);					
		}
	}
	return dwResult;
}

/*
 *	Blit solid color rectangle to surface
 *
 *	Parameters:		rcDestin	- destination rectangle to blit
 *					Color		- color (use RGB macro)
 *
 *	Return values:  DD_OK if success, else error code of Blit() function
 */
HRESULT CSurface::Blt(CRect rcDestin, COLORREF Color)
{
	HRESULT dwResult = 1; // 1 means: NOT INIT
	DDBLTFX fx;
	//
	// Check if the destination surface is valid
	if(IsValid()) {
		//
		// Fill DDBLTFX structure valid date
		fx.dwSize = sizeof(fx);
		fx.dwFillColor = Color;
		//
		// Blitting...
		while(1) {
			dwResult = GetSurface()->Blt(&rcDestin, NULL, NULL, DDBLT_COLORFILL, &fx);
			if(dwResult != DDERR_WASSTILLDRAWING) {
				break;
			}
		}
		if(DD_OK != dwResult) {	
			DXERR("Cannot blit solid rect due", dwResult);			
		}
	}
	return dwResult;
}

/*
 *		Copy bitmap to created surface
 *
 *		Parameters: csBitmap	-	full path of bitmap in storage file
 *
 *		Returns value:	DD_OK or ERROR_SUCCESS if success. Returns 1
 *						if the bitmap path is invalid (empty)
 *
 *		Note: While this function is processed, temporary file
 *		is created at disk. This file has been deleted
 *      after function ends. If bitmap is smaller or greater
 *		than surface, bitmap is stretched or shrinked.
 *		This function is casove narocna.
 */
HRESULT CSurface::CopyBitmap(CString csBitmap)
{
	HRESULT				dwReturn = 1;// = ERROR_NOINIT or BITMAP NAME IS INVALID;
	DWORD				dwSize;
	DWORD				dwWritten;
	LPBYTE				lpData;
	HANDLE				hFile = NULL;
	BITMAP				bm;
	HBITMAP				hBitmap;
	CString				strTempFile = _T("C:\\surface.tmp");
    HDC                 hdcImage;
    HDC                 hdc = NULL;
	HBITMAP				hOld;
	// 
	// If display initialized and storage opened
	if(!csBitmap.IsEmpty() && m_bInit) {
		//
		// Try to find file in storage
		dwReturn = stgGetFile3(csBitmap, &lpData, dwSize);
		if(dwReturn != ERROR_SUCCESS) {
			DXERR("Cannot get file from storage due:", dwReturn);
			return dwReturn;
		}
		//
		// Create temporary file
		hFile = (HANDLE)CreateFile(strTempFile, 
								GENERIC_WRITE , 
								FILE_SHARE_WRITE | FILE_SHARE_READ, 
								NULL, 
								CREATE_ALWAYS, 
								FILE_ATTRIBUTE_TEMPORARY, 
								NULL);
		if(hFile == INVALID_HANDLE_VALUE) {
			dwReturn = GetLastError();	
			DXERR("Cannot create temporary file due:", dwReturn);
			delete [] lpData;
			return dwReturn;
		}
		//
		// Write bitmap data to the temporary file
		WriteFile(hFile, lpData, dwSize, &dwWritten, NULL);	
		if(dwWritten != dwSize) {
			dwReturn = GetLastError();
			DXERR("Cannot write data due:", dwReturn);
			CloseHandle((HANDLE)hFile);	
			delete [] lpData;
			DeleteFile(strTempFile);
			return dwReturn;

		}
		//
		// Close it
		CloseHandle((HANDLE)hFile);								
		//
		// Get bitmap handle					
		// Try to load bitmap from file
		hBitmap = (HBITMAP)LoadImage(NULL, strTempFile, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE|LR_CREATEDIBSECTION);
		if (hBitmap == NULL) {
			dwReturn = GetLastError();
			DXERR("Cannot get bitmap handle due:", dwReturn);			
			delete [] lpData;
			DeleteFile(strTempFile);
			return dwReturn;
		}
		//
		// Get size of the bitmap
		if(!GetObject(hBitmap, sizeof(BITMAP), &bm))
		{
			dwReturn = GetLastError();
			DXERR("Cannot get bitmap information due:", dwReturn);			
			delete [] lpData;
			DeleteObject(hBitmap);
			DeleteFile(strTempFile);
			return dwReturn;
		}		
		//
		// Create DC for our bitmap
		hdcImage = CreateCompatibleDC(NULL);
		if (!hdcImage) {
			dwReturn = GetLastError();
			DXERR("Cannot create temporary DC due:", dwReturn);			
			delete [] lpData;
			DeleteObject(hBitmap);
			DeleteFile(strTempFile);
			return dwReturn;
		}
		//
		//  Select bitmap into a memoryDC so we can use it.
		hOld = (HBITMAP) SelectObject(hdcImage, hBitmap);
		//
		// Get DC of DD surface
		if ((dwReturn = m_lpDDSurface->GetDC(&hdc)) == DD_OK)
		{			
			//
			// Blit bitmap to GDI surface
			StretchBlt(hdc, 0, 0, Width(), Height(), hdcImage, 0, 0, bm.bmWidth , bm.bmHeight, SRCCOPY);
			if(DD_OK != (dwReturn = m_lpDDSurface->ReleaseDC(hdc))) 
			{
				DXERR("Cannot release surface DC due:", dwReturn);			
			}
			//
			// If we want to set color key, set it now else exit with OK
			if(m_bColorKey) {
				dwReturn = SetColorKey();																					
				if(dwReturn != ERROR_SUCCESS) {
					DXERR("Cannot set CK on surface due:", dwReturn);		

				}
			}		
			//
			//release handle to GDI object and handle to DC
			DeleteDC(hdcImage);	
			DeleteObject(hBitmap);
			DeleteFile(strTempFile);				
			delete [] lpData;

		}
		if(dwReturn != DD_OK) {			
			DXERR("Cannot get surface DC due:", dwReturn);			
			delete [] lpData;
			DeleteObject(hBitmap);
			DeleteDC(hdcImage);	
			DeleteFile(strTempFile);
			return dwReturn;
		}
	}
	return dwReturn;
}
