//---------------------------------------------------------------------------
//
//  DDFontEngine - Copyright  2000 Michael Ftsch.
//
//      The DirectDraw Bitmap Font Engine from "Fast Bitmap Fonts
//      for DirectDraw", which can be found at www.mr-gamemaker.com.
//      The tutorial and a TrueType-to-DDF converter are also available at
//      http://www.crosswinds.net/~foetsch
//
//---------------------------------------------------------------------------

#include "stdafx.h"

#define DISPLAY_API __declspec(dllexport)

#include "Common.h"
#include "DDFontEngine.h"


#define SAFERELEASE(x)      if (x) { x->Release(); x = NULL; }
#define SAFEDELETE(x)       if (x) { delete x; x = NULL; }
#define SAFEDELETEARRAY(x)  if (x) { delete[] x; x = NULL; }
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//-------------------------- class CDDFont ----------------------------------
//---------------------------------------------------------------------------

// CDDFont constructor
//      -> ddfe: Initialized CDDFontEngine that will be used to render the font
//      -> fontfile: Name of DDF file that should be loaded
//      -> textcolor: Desired color of text. Use TEXTCOLOR macro.
CDDFont::CDDFont()
{
    lpbi = NULL;
    lpBits = NULL;
}
//---------------------------------------------------------------------------

// CDDFont destructor
CDDFont::~CDDFont()
{    
    SAFEDELETEARRAY(lpBits);    // Free buffer for bitmap bits
    SAFEDELETEARRAY(lpbi);      // Free buffer for BITMAPINFO
}
//---------------------------------------------------------------------------

// CDDFont::CreateFont
//      Creates a DirectDraw Surface, loads the bitmap in the DDF File
//      into it and reads the text metrics and ABC widths into class structures
//
HRESULT CDDFont::CreateFont(CString csFontFile, CString csFontName, COLORREF crColor)
{
    HDC SurfDC=NULL;

	BYTE *lpData ,*lpBegin;
	DWORD dwSize, dwRet;
	DWORD dwPointer = 0;
	//
	// Fill in class
	m_csFont = csFontName;
	m_csFontFile = csFontFile;
	m_crColor = crColor;
	//
	// Try to get bitmap file from storage
	dwRet = stgGetFile3(csFontFile, &lpData, dwSize);
	if(dwRet != ERROR_SUCCESS) {
		DXTRACE6("Cannot create surface for font due", dwRet);
		return dwRet;
	}	
	lpBegin = lpData;

    BITMAPFILEHEADER bmfh;
    ZeroMemory(&bmfh, sizeof(bmfh));    
	memcpy(&bmfh, lpData, sizeof(bmfh));
	lpData += sizeof(bmfh);
	//
    // Check whether it's a valid DDF file:
    if ((bmfh.bfType != 19778) /*'BM'*/|| (bmfh.bfReserved1 != 'DD') || (bmfh.bfReserved2 != 'FF')) {
		DXTRACE0("File is not valid bitmap");
		return 1;
	}
    //
	// Read BITMAPINFO and bitmap bits:
    lpbi = (LPBITMAPINFO)(new char[bmfh.bfOffBits - sizeof(bmfh)]);
     //if (!lpbi) throw (int)2;    // ("new" should throw anyway but...)
	memcpy(lpbi, lpData, bmfh.bfOffBits-sizeof(bmfh));
	lpData += bmfh.bfOffBits-sizeof(bmfh);
    
    lpBits = (LPVOID)(new char[bmfh.bfSize-bmfh.bfOffBits]);
	memcpy(lpBits, lpData, bmfh.bfSize-bmfh.bfOffBits);
	lpData += bmfh.bfSize-bmfh.bfOffBits;
	//
	// Get Dimension of bitmap
    SurfWidth = lpbi->bmiHeader.biWidth;
    SurfHeight = lpbi->bmiHeader.biHeight;
	//
	// Create surface for font
	m_lpFontSurf.Create(SurfWidth, SurfHeight, 0);
	//
    // Read text metrics:
	memcpy(&TextMetrics, lpData, sizeof(TextMetrics));
	lpData += sizeof(TextMetrics);
	memcpy((LPVOID)((DWORD)ABCWidths+32*sizeof(ABC)), lpData,  224*sizeof(ABC));
	lpData += 224*sizeof(ABC);
	memcpy(&LogFont, lpData, sizeof(LogFont));
	lpData += sizeof(LogFont);
	//
	// Computes dim of one cell
	CellWidth = SurfWidth >> 4;
    CellHeight = SurfHeight / 14;

	
	// Pre-calculate SrcRects:
	ZeroMemory(SrcRects,	32*sizeof(RECT));
	ZeroMemory(ABCWidths,	32*sizeof(ABC));
	ZeroMemory(BPlusC,		32*sizeof(int));

	for (int c=32; c < 256; c++)
	{
		SrcRects[c].left = ((c-32) % 16) * CellWidth;
		SrcRects[c].top = ((c-32) >> 4) * CellHeight;
		SrcRects[c].right = SrcRects[c].left + ABCWidths[c].abcB;
		SrcRects[c].bottom = SrcRects[c].top + CellHeight;
		BPlusC[c] = ABCWidths[c].abcB + ABCWidths[c].abcC;
	}

	 DWORD *palentry = (DWORD*)&lpbi->bmiColors[1];
	 if(m_crColor != 0) {
		*palentry = m_crColor;
	 }

	
	dwRet = m_lpFontSurf.GetSurface()->GetDC(&SurfDC);
	if(dwRet != DD_OK) {
		DXTRACE6("Cannot Get DC due", dwRet);
		return dwRet;
	}
    StretchDIBits(	SurfDC,
					0, 0, SurfWidth, SurfHeight,
					0, 0, SurfWidth, SurfHeight,
					lpBits,
					lpbi,
					DIB_RGB_COLORS,
					SRCCOPY);
    m_lpFontSurf.GetSurface()->ReleaseDC(SurfDC);
	m_lpFontSurf.SetColorKey();

	delete [] lpBegin;

	return dwRet;
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//-------------------------- class CDDFontEngine ----------------------------
//---------------------------------------------------------------------------

CDDFontEngine::CDDFontEngine()
{
	m_pCurrentFont = NULL;
}
//---------------------------------------------------------------------------

CDDFontEngine::~CDDFontEngine()
{
	CDDFont *pFont;
	for(int i = 0; i < m_arFonts.GetSize() ; i++) {
		pFont = (CDDFont *) m_arFonts[i];
		delete pFont;
	}
}
//---------------------------------------------------------------------------

HRESULT CDDFontEngine::SelectFont(CString csFontName)
{
	CDDFont *pFont;
	for(int i = 0; i < m_arFonts.GetSize() ; i++) {
		pFont = (CDDFont *) m_arFonts[i];
		if(pFont->GetFontName() == csFontName) {
			m_pCurrentFont = pFont;
			return 0;
		}
	}
	return 1;    
}	
//---------------------------------------------------------------------------

HRESULT CDDFontEngine::DrawText(int x, int y,  CString csText, CString csFont)
{
    CRect destin;
	UCHAR ch;

	if(ERROR_SUCCESS != SelectFont(csFont)) {
		DXTRACE0("Specified font is not defined");
		return 1;
	}

    for (int i = 0; i < csText.GetLength(); i++)
    {
        ch = csText[i];
        x += m_pCurrentFont->ABCWidths[ch].abcA;
		destin.left = x;
		destin.top  = y;
		destin.right = destin.left + (m_pCurrentFont->SrcRects[ch].right - m_pCurrentFont->SrcRects[ch].left);
		destin.bottom  = destin.top + (m_pCurrentFont->SrcRects[ch].bottom - m_pCurrentFont->SrcRects[ch].top);
		disBlt(destin, m_pCurrentFont->SrcRects[ch], &m_pCurrentFont->m_lpFontSurf);
        x += m_pCurrentFont->BPlusC[ch];
    }
    return DD_OK;
}
//---------------------------------------------------------------------------
HRESULT	CDDFontEngine::CreateFont(CString csFontFile, CString csFontName, COLORREF crColor)
{
	DWORD dwRet;
	CDDFont *pNewFont = new CDDFont;
	dwRet = pNewFont->CreateFont(csFontFile, csFontName, crColor);
	if(dwRet != ERROR_SUCCESS) {
		DXTRACE6("Cannot create new font due", dwRet);
		return dwRet;
	}
	m_arFonts.Add(pNewFont);
	
	return dwRet;
}


